# -----------------------------------------------------------
# "Image-Text Alignment using Adaptive Cross-attention with Transformer Encoder for Scene Graphs"
# Juyong Song and Sunghyun Choi
#
# Modified by Juyong Song, 2021
# ---------------------------------------------------------------

# -----------------------------------------------------------
# "Cross-modal Scene Graph Matching for Relationship-aware Image-Text Retrieval"
# Sijin Wang, Ruiping Wang, Ziwei yao, Shiguang Shan and Xilin Chen
# https://arxiv.org/abs/1910.05134
#
# Written by Sijin Wang, 2019
# ---------------------------------------------------------------

# -----------------------------------------------------------
# Stacked Cross Attention Network implementation based on
# https://arxiv.org/abs/1803.08024.
# "Stacked Cross Attention for Image-Text Matching"
# Kuang-Huei Lee, Xi Chen, Gang Hua, Houdong Hu, Xiaodong He
#
# Written by Kuang-Huei Lee, 2018
# ---------------------------------------------------------------

import time
import torch
import torch.utils.data as data
import os
import numpy as np
import json
import pickle

from util import get_enc_tokenizer, get_sep_token_id


class PrecompDataset(data.Dataset):
    """
    Load precomputed captions and image features
    """

    def __init__(self, data_path, data_split, image_sg_file, caption_data_file, feature_name, opt):
        self.opt = opt
        self.tokenizer = get_enc_tokenizer(opt.enc_type)
        self.is_triplet = opt.is_triplet
        self.only_cap = opt.only_cap
        ts = time.time()
        # Captions
        self.cache_dir = os.path.join(opt.cache_path, opt.data_name, opt.enc_type)
        print('caption data is loading....')
        try:
            with open(os.path.join(self.cache_dir, data_split + '_captions.pickle'), "rb") as f:
                self.captions = pickle.load(f)
            with open(os.path.join(self.cache_dir, data_split + '_cap_lens.pickle'), "rb") as f:
                self.cap_lens = pickle.load(f)
            with open(os.path.join(self.cache_dir, data_split + '_caption_rel.pickle'), "rb") as f:
                self.caption_rel = pickle.load(f)
            with open(os.path.join(self.cache_dir, data_split + '_cap_rel_nums.pickle'), "rb") as f:
                self.cap_rel_nums = pickle.load(f)
        except(OSError, IOError) as e:

            self.captions = []
            self.cap_lens = []
            self.caption_rel = []
            self.cap_rel_nums = []

            if 'coco' in caption_data_file:
                caption_key = 'caption'
                rel_key = 'relations'
            elif 'flickr' in caption_data_file:
                caption_key = 'sent'
                rel_key = 'rels'
            else:
                raise ValueError("caption_data_file should be contain 'coco' or 'flickr'")

            with open(os.path.join(data_path, caption_data_file), 'rb') as f:
                caption_data = json.load(f)
                for ann in caption_data:
                    for cap in ann['captions'][:5]:
                        self.captions.append(cap[caption_key])
                        tokens = self.tokenizer(cap[caption_key])
                        self.cap_lens.append(len(tokens['input_ids']))

                        rel_concat = ''
                        for rel in cap[rel_key]:
                            if len(rel) == 3:
                                if len(rel_concat):
                                    rel_concat += '[SEP]'
                                rel_concat += rel[0] + ' ' + rel[2] + ' ' + rel[1]
                        self.caption_rel.append(rel_concat)
                        tokens = self.tokenizer(rel_concat, return_tensors='pt', padding=True)
                        where_sep_tokens = (tokens.input_ids == get_sep_token_id(opt.enc_type)).nonzero(
                            as_tuple=False).long()
                        self.cap_rel_nums.append(len(where_sep_tokens))

            os.makedirs(self.cache_dir, exist_ok=True)
            with open(os.path.join(self.cache_dir, data_split + '_captions.pickle'), "wb") as f:
                pickle.dump(self.captions, f)
            with open(os.path.join(self.cache_dir, data_split + '_cap_lens.pickle'), "wb") as f:
                pickle.dump(self.cap_lens, f)
            with open(os.path.join(self.cache_dir, data_split + '_caption_rel.pickle'), "wb") as f:
                pickle.dump(self.caption_rel, f)
            with open(os.path.join(self.cache_dir, data_split + '_cap_rel_nums.pickle'), "wb") as f:
                pickle.dump(self.cap_rel_nums, f)

        print('caption data loading takes {:.4f} secs'.format(time.time() - ts))
        # sg information
        with open(os.path.join(data_path, image_sg_file), 'r') as f:
            sg_data = json.load(f)
            self.obj_vocab = sg_data['object_dict']
            self.rel_vocab = sg_data['predicate_dict']
            self.object_num = sg_data['object_num']
            self.predicate_num = sg_data['predicate_num']
            img_to_sg = sg_data['img_to_sg']

            self.obj_num_per_img = [
                sg['sg']['object'].index(-1) if -1 in sg['sg']['object'] else len(sg['sg']['object']) for sg in
                img_to_sg]
            # self.obj_num_per_img = [min(36, num) for num in self.obj_num_per_img]
            self.predicate_num_per_img = [
                sg['sg']['relationship'].index([-1, -1, -1]) if [-1, -1, -1] in sg['sg']['relationship'] else len(
                    sg['sg']['relationship']) for sg in img_to_sg]

            self.obj_list_per_img = []
            self.rel_list_per_img = []
            self.obj_label_list_per_img = []
            self.rel_label_list_per_img = []
            for i, sg in enumerate(img_to_sg):
                obj_num = self.obj_num_per_img[i]
                obj_list = sg['sg']['object'][:obj_num]
                obj_label_list = [self.obj_vocab[ob] for ob in obj_list]

                self.obj_list_per_img.append(obj_list)  # (N, obj_num=36)
                self.obj_label_list_per_img.append(obj_label_list)

                pred_num = self.predicate_num_per_img[i]
                rel_list = sg['sg']['relationship'][:pred_num]
                if self.is_triplet:
                    rel_label_list = [obj_label_list[s] + ' ' + self.rel_vocab[p] + ' ' + obj_label_list[o] for s, p, o
                                      in rel_list]
                else:
                    rel_label_list = [self.rel_vocab[p] for _, p, _ in rel_list]

                self.rel_list_per_img.append(rel_list)  # (N, pred_num=25, 3)
                self.rel_label_list_per_img.append(rel_label_list)

        self.image_feature_root = os.path.join(data_path, 'npys', 'obj_' + data_split)
        self.image_predicate_feature_root = os.path.join(data_path, 'npys', 'rel_' + data_split)

        self.length = len(self.captions)
        self.im_div = 5

        if data_split == 'val':
            self.length = 5000

        print(f'data length : {self.length}')

    def __getitem__(self, index):
        # handle the image redundancy
        # Get obj features of an image
        img_id = index // self.im_div
        obj_num = self.obj_num_per_img[img_id]
        image = np.load(os.path.join(self.image_feature_root, str(img_id) + '.npy'))[
                :obj_num]  # (36, img_dim) (img_dim = 2048?)
        image = torch.from_numpy(image)
        obj_list = self.obj_list_per_img[img_id]  # (obj_num=36,)
        # Object Label List
        obj_label_list = self.obj_label_list_per_img[img_id]
        if not self.opt.no_rel:
            # Get relation features and rel list of an image
            pred_num = self.predicate_num_per_img[img_id]
            image_predicate = np.load(os.path.join(self.image_predicate_feature_root, str(img_id) + '.npy'))[
                              :pred_num]  # (25, img_dim)
            image_predicate = torch.from_numpy(image_predicate)
            rel_list = self.rel_list_per_img[img_id]  # (pred_num=25, 3)
            # Relation Label List
            rel_label_list = self.rel_label_list_per_img[img_id]
        else:
            pred_num = None
            image_predicate = None
            rel_list = None
            rel_label_list = None

        # Get the corresponding caption
        caption = self.captions[index]  # A sentence
        length = self.cap_lens[index]

        if not self.opt.no_rel:
            cap_rel = caption if self.only_cap else self.caption_rel[index]
            cap_rel_num = length if self.only_cap else self.cap_rel_nums[index]
        else:
            cap_rel = None
            cap_rel_num = None

        return image, obj_num, caption, index, img_id, image_predicate, pred_num, rel_list, obj_list, cap_rel, \
               cap_rel_num, length, obj_label_list, rel_label_list, self.opt

    def __len__(self):
        return self.length


def collate_fn(data):
    # Sort a data list by caption length
    data.sort(key=lambda x: len(x[2]), reverse=True)
    images, obj_nums, captions, ids, img_ids, image_predicates, pred_nums, rel_lists, obj_lists, cap_rels, \
    cap_rel_nums, lengths, obj_label_lists, rel_label_lists, opt = zip(*data)
    # Renumber all object in a batch obj_lists : (batch, obj_num=36)
    objs = []
    obj_labels = []
    for obj_list, obj_label_list in zip(obj_lists, obj_label_lists):
        # obj_list : (obj_num=36,)
        objs = objs + obj_list
        obj_labels = obj_labels + obj_label_list

    objs = torch.tensor(objs, dtype=torch.int64)  # (batch_size*obj_num, )

    # image : (batch_size, obj_num, dim)
    images = torch.cat(images, 0)  # (batch_size*obj_num, dim)

    if opt[0].no_rel:
        rels = None
        rel_labels = None
    else:
        obj_offset = 0
        rels = []
        for i, (rel_list) in enumerate(rel_lists):
            # rel_list : (pred_num, 3)
            obj_num = obj_nums[i]
            for s, p, o in rel_list:
                rels.append([s + obj_offset, p, o + obj_offset])
            obj_offset += obj_num
        rels = torch.tensor(rels, dtype=torch.int64)  # (batch_size*pred_num, 3)
        # image_predicates : (batch_size, pred_num, dim)
        image_predicates = torch.cat(image_predicates, 0)

        assert len(rels) == len(image_predicates)

        rel_labels = []
        for rel_label_list in rel_label_lists:
            for rel_label in rel_label_list:
                rel_labels.append(rel_label)

    return images, obj_nums, captions, lengths, ids, image_predicates, pred_nums, rels, objs, cap_rels, cap_rel_nums, \
           obj_labels, rel_labels


def get_precomp_loader(data_path, data_split, file_list, feature_name, opt, batch_size=100,
                       shuffle=True, num_workers=2):
    dset = PrecompDataset(data_path, data_split, file_list[0], file_list[1], feature_name, opt=opt)
    if (opt.part_of_data is not None) and (data_split == 'train'):
        indices = torch.randperm(len(dset))[:int(len(dset) * opt.part_of_data)].tolist()
        dset = torch.utils.data.Subset(dset, indices)

    data_loader = torch.utils.data.DataLoader(dataset=dset,
                                              batch_size=batch_size,
                                              shuffle=shuffle,
                                              num_workers=num_workers,
                                              pin_memory=True,
                                              collate_fn=collate_fn)
    return data_loader


def get_loaders(data_name, train_file_list, val_file_list, feature_name, batch_size, workers, opt):
    dpath = opt.data_path
    if opt.debug:
        train_loader = get_precomp_loader(dpath, 'val', val_file_list, feature_name, opt, batch_size, True, workers)
    else:
        train_loader = get_precomp_loader(dpath, 'train', train_file_list, feature_name, opt, batch_size, True, workers)
    val_loader = get_precomp_loader(dpath, 'val', val_file_list, feature_name, opt, batch_size, False, workers)

    return train_loader, val_loader


def get_test_loader(split_name, data_name, test_file_list, feature_name, batch_size,
                    workers, opt):
    dpath = opt.data_path
    print(dpath, split_name, data_name, test_file_list, feature_name)
    test_loader = get_precomp_loader(dpath, split_name, test_file_list, feature_name, opt, batch_size, False, workers)
    return test_loader
