# -----------------------------------------------------------
# "Image-Text Alignment using Adaptive Cross-attention with Transformer Encoder for Scene Graphs"
# Juyong Song and Sunghyun Choi
#
# Modified by Juyong Song, 2021
# ---------------------------------------------------------------

# -----------------------------------------------------------
# "Cross-modal Scene Graph Matching for Relationship-aware Image-Text Retrieval"
# Sijin Wang, Ruiping Wang, Ziwei yao, Shiguang Shan and Xilin Chen
# https://arxiv.org/abs/1910.05134
#
# Written by Sijin Wang, 2019
# ---------------------------------------------------------------

# -----------------------------------------------------------
# Stacked Cross Attention Network implementation based on
# https://arxiv.org/abs/1803.08024.
# "Stacked Cross Attention for Image-Text Matching"
# Kuang-Huei Lee, Xi Chen, Gang Hua, Houdong Hu, Xiaodong He
#
# Written by Kuang-Huei Lee, 2018
# ---------------------------------------------------------------

"""Evaluation"""

from __future__ import print_function

import time
from collections import OrderedDict

import numpy as np
import torch

from data import get_test_loader
from model import RELAX


class AverageMeter(object):
    """Computes and stores the average and current value"""

    def __init__(self):
        self.reset()

    def reset(self):
        self.val = 0
        self.avg = 0
        self.sum = 0
        self.count = 0

    def update(self, val, n=0):
        self.val = val
        self.sum += val * n
        self.count += n
        self.avg = self.sum / (.0001 + self.count)

    def __str__(self):
        """String representation for logging
        """
        # for values that should be recorded exactly e.g. iteration number
        if self.count == 0:
            return str(self.val)
        # for stats
        return '%.4f (%.4f)' % (self.val, self.avg)


class LogCollector(object):
    """A collection of logging objects that can change from train to val"""

    def __init__(self):
        # to keep the order of logged variables deterministic
        self.meters = OrderedDict()

    def update(self, k, v, n=0):
        # create a new meter if previously not recorded
        if k not in self.meters:
            self.meters[k] = AverageMeter()
        self.meters[k].update(v, n)

    def __str__(self):
        """Concatenate the meters in one log line
        """
        s = ''
        for i, (k, v) in enumerate(self.meters.items()):
            if i > 0:
                s += '  '
            s += k + ' ' + str(v)
        return s

    def tb_log(self, tb_logger, prefix='', step=None):
        """Log using tensorboard
        """
        for k, v in self.meters.items():
            tb_logger.log_value(prefix + k, v.val, step=step)


def encode_data_with_predicate(model, data_loader, opt, log_step=10, logging=print):
    """Encode all images and captions loadable by `data_loader`
    """
    batch_time = AverageMeter()
    val_logger = LogCollector()

    # switch to evaluate mode
    model.eval()  # val_start()
    end = time.time()

    # np array to keep all the embeddings
    img_embs = None
    cap_embs = None
    cap_lens = None
    obj_nums = None
    pred_embs = None
    pred_nums = None
    cap_rel_embs = None
    cap_rel_nums = None

    max_n_word = 0
    max_n_cap_rel = 0

    for i, (images, obj_num, captions, lengths, ids, image_predicates, pred_num, rels, objs, cap_rels,
            cap_rel_num, obj_labels, rel_labels) in enumerate(data_loader):
        max_n_word = max(max_n_word, max(lengths))
        if not opt.no_rel:
            max_n_cap_rel = max(max_n_cap_rel, max(cap_rel_num)+1)

    for i, (images, obj_num, captions, lengths, ids, image_predicates, pred_num,
            rels, objs, cap_rels, cap_rel_num, obj_labels, rel_labels) in enumerate(data_loader):

        # make sure val logger is used
        model.logger = val_logger
        img_emb, cap_emb, cap_len, pred_emb, cap_rel_emb, cap_rel_num = \
                model(images, obj_num, captions, lengths, image_predicates,
                    pred_num, rels, objs, cap_rels, cap_rel_num, obj_labels, rel_labels)

        if img_embs is None:
            if img_emb.dim() == 3:
                img_embs = np.zeros((len(data_loader.dataset), img_emb.size(1), img_emb.size(2)))
                if not opt.no_rel:
                    pred_embs = np.zeros((len(data_loader.dataset), pred_emb.size(1), pred_emb.size(2)))
            else:
                img_embs = np.zeros((len(data_loader.dataset), img_emb.size(1)))
                if not opt.no_rel:
                    pred_embs = np.zeros((len(data_loader.dataset), pred_emb.size(1)))
            obj_nums = [0] * len(data_loader.dataset)
            cap_embs = np.zeros((len(data_loader.dataset), max_n_word, cap_emb.size(2)))
            cap_lens = [0] * len(data_loader.dataset)
            if not opt.no_rel:
                pred_nums = [0] * len(data_loader.dataset)
                cap_rel_embs = np.zeros((len(data_loader.dataset), max_n_cap_rel, cap_rel_emb.size(2)))
                cap_rel_nums = [0] * len(data_loader.dataset)

        # cache embeddings
        img_embs[ids] = img_emb.data.cpu().numpy().copy()
        cap_embs[ids, :max(lengths), :] = cap_emb.data.cpu().numpy().copy()
        for j, nid in enumerate(ids):
            cap_lens[nid] = cap_len[j]
            obj_nums[nid] = obj_num[j]
        if not opt.no_rel:
            pred_embs[ids] = pred_emb.data.cpu().numpy().copy()
            cap_rel_embs[ids, :max(cap_rel_num), :] = cap_rel_emb.data.cpu().numpy().copy()
            for j, nid in enumerate(ids):
                pred_nums[nid] = pred_num[j]
                cap_rel_nums[nid] = cap_rel_num[j]

        glb_img, glb_txt = model.get_global(img_emb, cap_emb, model.adapt_img, model.adapt_txt, use_cxt=False)
        if not opt.no_rel:
            glb_rel_img, glb_rel_txt = model.get_global(pred_emb, cap_rel_emb, model.rel_adapt_img, model.rel_adapt_txt, use_cxt=True)

        scores1 = model.similarity(img_emb, cap_emb, cap_lens,
                                   glb_img=glb_img, glb_txt=glb_txt, exclude_last=True)
        if not opt.no_rel:
            scores2 = model.rel_sim(pred_emb, cap_rel_emb, cap_rel_nums,
                                glb_img=glb_rel_img, glb_txt=glb_rel_txt)
        else:
            scores2 = torch.zeros_like(scores1)
        scores = scores1 + opt.predicate_score_rate * scores2

        if 'mi' in opt.loss_type:
            scores *= model.inv_t
        model.forward_loss(scores)

        # measure elapsed time
        batch_time.update(time.time() - end)
        end = time.time()

        if i % log_step == 0:
            logging('Test: [{0}/{1}]\t'
                    '{e_log}\t'
                    'Time {batch_time.val:.3f} ({batch_time.avg:.3f})\t'
                .format(
                i, len(data_loader), batch_time=batch_time,
                e_log=str(model.logger)))
        del images, captions, rels, cap_rels, image_predicates
    return img_embs, obj_nums, cap_embs, cap_lens, pred_embs, pred_nums, cap_rel_embs, cap_rel_nums


def evalrank(model_path, data_path=None, split='dev', fold5=False):
    """
    Evaluate a trained model on either dev or test. If `fold5=True`, 5 fold
    cross-validation is done (only for MSCOCO). Otherwise, the full data is
    used for evaluation.
    """
    # load model and options
    checkpoint = torch.load(model_path)
    opt = checkpoint['opt']
    if data_path is not None:
        opt.data_path = data_path

    print(opt)
    # construct model
    model = RELAX(opt)
    # load model state
    model.load_state_dict(checkpoint['model'])
    model.inv_t = torch.zeros(1).fill_(10).cuda()
    # opt.batch_size = 10  # for local evaluation
    opt.workers = 0

    print('Loading dataset')
    test_file_list = [opt.data_name.replace('split', split) + opt.sg_file_name,
                      opt.data_name.replace('split', split) + opt.caption_data_file_name]
    print("TEST FILE LIST : ", test_file_list)
    data_loader = get_test_loader(split, opt.data_name, test_file_list, opt.feature_name,
                                  opt.batch_size, opt.workers, opt)

    with torch.no_grad():
        print('Computing results...')
        img_embs, obj_nums, cap_embs, cap_lens, pred_embs, pred_nums, cap_rel_embs, cap_rel_nums \
            = encode_data_with_predicate(model, data_loader, opt)
        print('Images: %d, Captions: %d' %
              (img_embs.shape[0] / 5, cap_embs.shape[0]))

        # if not fold5:
        # no cross-validation, full evaluation
        img_embs_shared = torch.FloatTensor([img_embs[i] for i in range(0, len(img_embs), 5)]).cuda()
        obj_nums_shared = [obj_nums[i] for i in range(0, len(obj_nums), 5)]
        cap_embs_shared = torch.FloatTensor(cap_embs).cuda()

        if not opt.no_rel:
            pred_embs_shared = torch.FloatTensor([pred_embs[i] for i in range(0, len(pred_embs), 5)]).cuda()
            pred_nums_shared = [pred_nums[i] for i in range(0, len(pred_nums), 5)]    
            cap_rel_embs_shared = torch.FloatTensor(cap_rel_embs).cuda()
    
        start = time.time()

        glb_img, glb_txt = model.get_global(img_embs_shared, cap_embs_shared,
                model.adapt_img, model.adapt_txt, use_cxt=False)
        if not opt.no_rel:
            glb_rel_img, glb_rel_txt = model.get_global(pred_embs_shared, cap_rel_embs_shared,
                model.rel_adapt_img, model.rel_adapt_txt, use_cxt=True)
        sims1 = model.similarity.forward_shared(img_embs_shared, cap_embs_shared, cap_lens,
                                                glb_img=glb_img, glb_txt=glb_txt,
                                                shared_size=128, exclude_last=True)
        if not opt.no_rel:
            sims2 = model.rel_sim.forward_shared(pred_embs_shared, cap_rel_embs_shared, cap_rel_nums,
                                             glb_img=glb_rel_img, glb_txt=glb_rel_txt,
                                             shared_size=128)
        else:
            sims2 = torch.zeros_like(sims1)
        sims = sims1 + opt.predicate_score_rate * sims2
        end = time.time()

        print("calculate similarity time:", end - start)

        sims = sims.data.cpu().numpy()
        r, rt, i2t_results = i2t(img_embs, cap_embs, cap_lens, sims, return_ranks=True)
        ri, rti, t2i_results = t2i(img_embs, cap_embs, cap_lens, sims, return_ranks=True)
        ar = (r[0] + r[1] + r[2]) / 3
        ari = (ri[0] + ri[1] + ri[2]) / 3
        rsum = r[0] + r[1] + r[2] + ri[0] + ri[1] + ri[2]
        print("rsum: %.1f" % rsum)
        print("Average i2t Recall: %.1f" % ar)
        print("Image to text: %.1f %.1f %.1f %.1f %.1f" % r)
        print("Average t2i Recall: %.1f" % ari)
        print("Text to image: %.1f %.1f %.1f %.1f %.1f" % ri)
        results = [rsum] + list(r)[:3] + [ar] + list(ri)[:3] + [ari]
        print(results)

        if fold5:
            # 5fold cross-validation, only for MSCOCO
            results_fold5 = []
            i2t_results = []
            t2i_results = []
            sims_fold5 = []

            for i in range(5):
                start = time.time()

                img_embs_shard = img_embs[i * 5000:(i + 1) * 5000:5]
                obj_nums_shard = obj_nums[i * 5000:(i + 1) * 5000:5]
                cap_embs_shard = cap_embs[i * 5000:(i + 1) * 5000]
                cap_lens_shard = cap_lens[i * 5000:(i + 1) * 5000]
                img_embs_shard = torch.FloatTensor(img_embs_shard).cuda()
                cap_embs_shard = torch.FloatTensor(cap_embs_shard).cuda()
                if not opt.no_rel:
                    pred_embs_shard = pred_embs[i * 5000:(i + 1) * 5000:5]
                    pred_nums_shard = pred_nums[i * 5000:(i + 1) * 5000:5]
                    cap_rel_embs_shard = cap_rel_embs[i * 5000:(i + 1) * 5000]
                    cap_rel_nums_shard = cap_rel_nums[i * 5000:(i + 1) * 5000]
                    pred_embs_shard = torch.FloatTensor(pred_embs_shard).cuda()
                    cap_rel_embs_shard = torch.FloatTensor(cap_rel_embs_shard).cuda()

                glb_img, glb_txt = model.get_global(img_embs_shard, cap_embs_shard,
                                model.adapt_img, model.adapt_txt, use_cxt=False)

                sims1 = model.similarity.forward_shared(img_embs_shard, cap_embs_shard, cap_lens_shard,
                                                        glb_img=glb_img, glb_txt=glb_txt,
                                                        shared_size=128, exclude_last=True)
                if not opt.no_rel:
                    glb_rel_img, glb_rel_txt = model.get_global(pred_embs_shard, cap_rel_embs_shard,
                                    model.rel_adapt_img, model.rel_adapt_txt, use_cxt=True)    
                    sims2 = model.rel_sim.forward_shared(pred_embs_shard, cap_rel_embs_shard, cap_rel_nums_shard,
                                                     glb_img=glb_rel_img, glb_txt=glb_rel_txt, shared_size=128)
                else:
                    sims2 = torch.zeros_like(sims1)
                sims_shared = sims1 + opt.predicate_score_rate * sims2

                end = time.time()
                print("calculate similarity time:", end - start)

                sims_shared = sims_shared.data.cpu().numpy()
                r, rt0, r_results = i2t(img_embs_shard, cap_embs_shard, cap_lens_shard, sims_shared, return_ranks=True)
                print("Image to text: %.1f, %.1f, %.1f, %.1f, %.1f" % r)
                ri, rti0, ri_results = t2i(img_embs_shard, cap_embs_shard, cap_lens_shard, sims_shared, return_ranks=True)
                print("Text to image: %.1f, %.1f, %.1f, %.1f, %.1f" % ri)
                i2t_results.append(r_results)
                t2i_results.append(ri_results)
                if i == 0:
                    rt, rti = rt0, rti0
                ar = (r[0] + r[1] + r[2]) / 3
                ari = (ri[0] + ri[1] + ri[2]) / 3
                rsum = r[0] + r[1] + r[2] + ri[0] + ri[1] + ri[2]
                print("rsum: %.1f ar: %.1f ari: %.1f" % (rsum, ar, ari))
                results_fold5 += [list(r) + list(ri) + [ar, ari, rsum]]

                sims_fold5.append(sims_shared.tolist())

            print("-----------------------------------")
            print("Mean metrics: ")
            mean_metrics = tuple(np.array(results_fold5).mean(axis=0).flatten())
            print("rsum: %.1f" % (mean_metrics[10] * 6))
            print("Average i2t Recall: %.1f" % mean_metrics[11])
            print("Image to text: %.1f %.1f %.1f %.1f %.1f" %
                  mean_metrics[:5])
            print("Average t2i Recall: %.1f" % mean_metrics[12])
            print("Text to image: %.1f %.1f %.1f %.1f %.1f" %
                  mean_metrics[5:10])

    if fold5:
        return sims, results, sims_fold5, results_fold5
    else:
        return sims, results


def i2t(images, captions, caplens, sims, npts=None, return_ranks=False):
    npts = sims.shape[0]
    ranks = np.zeros(npts)
    top1 = np.zeros(npts)
    results = []
    for index in range(npts):
        result = dict()
        result['id'] = index
        inds = np.argsort(sims[index])[::-1]
        result['top5'] = list(inds[:5].tolist())
        result['top1'] = int(inds[0])
        result['top10'] = list(inds[:10].tolist())
        result['ranks'] = []
        # Score
        rank = 1e20
        for i in range(5 * index, 5 * index + 5, 1):
            tmp = int(np.where(inds == i)[0][0])
            result['ranks'].append((i, tmp))
            if tmp < rank:
                rank = tmp
        ranks[index] = int(rank)
        top1[index] = int(inds[0])

        if rank < 1:
            result['is_top1'] = 1
        else:
            result['is_top1'] = 0
        if rank < 5:
            result['is_top5'] = 1
        else:
            result['is_top5'] = 0

        results.append(result)

    # Compute metrics
    r1 = 100.0 * len(np.where(ranks < 1)[0]) / len(ranks)
    r5 = 100.0 * len(np.where(ranks < 5)[0]) / len(ranks)
    r10 = 100.0 * len(np.where(ranks < 10)[0]) / len(ranks)
    medr = float(np.floor(np.median(ranks)) + 1)
    meanr = float(ranks.mean() + 1)
    if return_ranks:
        return (r1, r5, r10, medr, meanr), (ranks, top1), results
    else:
        return (r1, r5, r10, medr, meanr), results


def t2i(images, captions, caplens, sims, npts=None, return_ranks=False):
    npts = sims.shape[0]
    ranks = np.zeros(5 * npts)
    top1 = np.zeros(5 * npts)

    # --> (5N(caption), N(image))
    sims = sims.T
    results = []
    for index in range(npts):
        for i in range(5):
            result = dict()
            result['id'] = 5 * index + i
            inds = np.argsort(sims[5 * index + i])[::-1]
            result['top5'] = list(inds[:5].tolist())
            result['top10'] = list(inds[:10].tolist())
            result['top1'] = int(inds[0])
            ranks[5 * index + i] = int(np.where(inds == index)[0][0])
            top1[5 * index + i] = int(inds[0])

            if ranks[5 * index + i] < 1:
                result['is_top1'] = 1
            else:
                result['is_top1'] = 0

            if ranks[5 * index + i] < 5:
                result['is_top5'] = 1
            else:
                result['is_top5'] = 0
            result['ranks'] = [(index, ranks[5 * index + i])]
            results.append(result)

    # Compute metrics
    r1 = 100.0 * len(np.where(ranks < 1)[0]) / len(ranks)
    r5 = 100.0 * len(np.where(ranks < 5)[0]) / len(ranks)
    r10 = 100.0 * len(np.where(ranks < 10)[0]) / len(ranks)
    medr = float(np.floor(np.median(ranks)) + 1)
    meanr = float(ranks.mean() + 1)
    if return_ranks:
        return (r1, r5, r10, medr, meanr), (ranks, top1), results
    else:
        return (r1, r5, r10, medr, meanr), results
