# -----------------------------------------------------------
# "Image-Text Alignment using Adaptive Cross-attention with Transformer Encoder for Scene Graphs"
# Juyong Song and Sunghyun Choi
#
# Modified by Juyong Song, 2021
# ---------------------------------------------------------------

# -----------------------------------------------------------
# "Cross-modal Scene Graph Matching for Relationship-aware Image-Text Retrieval"
# Sijin Wang, Ruiping Wang, Ziwei yao, Shiguang Shan and Xilin Chen
# https://arxiv.org/abs/1910.05134
#
# Written by Sijin Wang, 2019
# ---------------------------------------------------------------

# -----------------------------------------------------------
# Stacked Cross Attention Network implementation based on
# https://arxiv.org/abs/1803.08024.
# "Stacked Cross Attention for Image-Text Matching"
# Kuang-Huei Lee, Xi Chen, Gang Hua, Houdong Hu, Xiaodong He
#
# Written by Kuang-Huei Lee, 2018
# ---------------------------------------------------------------

import sys

import numpy as np
import torch
import torch.backends.cudnn as cudnn
import torch.nn as nn
import torch.nn.functional as F
import torch.nn.init
from torch.nn.utils.clip_grad import clip_grad_norm_ as clip_grad_norm

from graph import GraphTripleConv, GraphTripleConvNet, FusionLayer
from util import get_enc_tokenizer, get_enc_model, get_sep_token_id


# Because the authors of ADAPT(AAAI20) do not allow us to use their codes,
# we do not upload the codes for adaptation.


class Similarity(nn.Module):

    def __init__(self, similarity_object, opt, sim_name, max_attn):
        super().__init__()
        self.opt = opt
        self.similarity = similarity_object(opt.embed_size, sim_name, opt, max_attn)

        print(f'Created similarity: {similarity_object} {sim_name}')

    def forward(self, img_embed, cap_embed, lens, glb_img=None, glb_txt=None,
                shared=False, exclude_last=False):
        return self.similarity(img_embed, cap_embed, lens,
                               glb_img=glb_img, glb_txt=glb_txt, exclude_last=exclude_last)

    def forward_shared(self, img_embed, cap_embed, lens, glb_img=None, glb_txt=None,
                       shared_size=128, exclude_last=False):
        """
        Compute pairwise i2t image-caption distance with locality sharding
        """
        n_im_shard = (len(img_embed)-1)//shared_size + 1
        n_cap_shard = (len(cap_embed)-1)//shared_size + 1

        d = torch.zeros(len(img_embed), len(cap_embed)).cpu()
        for i in range(n_im_shard):
            im_start = shared_size*i
            im_end = min(shared_size*(i+1), len(img_embed))
            for j in range(n_cap_shard):
                sys.stdout.write('\r>> shard_xattn_t2i batch (%d,%d)' % (i, j))
                cap_start = shared_size*j
                cap_end = min(shared_size*(j+1), len(cap_embed))
                im = img_embed[im_start:im_end]
                s = cap_embed[cap_start:cap_end]
                l = lens[cap_start:cap_end]
                if glb_img is not None:
                    img_g = glb_img[im_start:im_end]
                else:
                    img_g = None
                if glb_txt is not None:
                    txt_g = glb_txt[cap_start:cap_end]
                else:
                    txt_g = None
                sim = self.forward(im, s, l, img_g, txt_g, exclude_last=exclude_last)
                d[im_start:im_end, cap_start:cap_end] = sim
        sys.stdout.write('\n')
        return d


class AdaptiveEmbedding(nn.Module):
    def __init__(self, latent_size, sim_name, opt, max_attn=False):
        super().__init__()
        self.sim_name = sim_name
        # self.adapt_img = ADAPT(latent_size)
        # self.adapt_txt = ADAPT(latent_size)
        self.max_attn = max_attn
        self.opt = opt

    def forward(self, images, captions, cap_lens,
                glb_img=None, glb_txt=None, exclude_last=False):
        """
        images : (B, L, D)
        (removed) obj_nums : the number of objects while L is the maximum
        captions : (C, T, D)
        cap_lens : the number of words while T is the maximum

        Because the number of image objects are almost 36,
        we will reduce the lengths of the word from T to cap_lens[i],
        rather than reducing the lengths of the objects from L to obj_nums[i].

        For T2I, query (Q) is the text,
        key(K) and value (V) is the images. (imgnorm seems good before adaptive embedding)

        For I2T, query (Q) is the images, (imgnorm seems bad)
        key(K) and value (V) is the text.

        glb_img : Global embedding of the images, for example, average pooling.
        (B, D)
        glb_txt : Global embedding of the text, for example, [CLS] embedding of BERT.
        (B, D)

        Thus, for the attention, the softmax should be executed following the imgs.
        """
        similarities = []
        n_image = images.size(0)
        n_caption = captions.size(0)

        for i in range(n_caption):
            if 't2i' in self.sim_name:
                cap_i = captions[i, :cap_lens[i]].unsqueeze(0)
            elif 'i2t' in self.sim_name:
                if exclude_last:
                    cap_i = captions[i, 1:cap_lens[i] - 1].unsqueeze(0)
                else:
                    cap_i = captions[i, 1:cap_lens[i]].unsqueeze(0)

            cap_i = cap_i.repeat(n_image, 1, 1)  # (B, T, D)

            # if glb_img is not None:
            #     cap_t = self.adapt_txt(value=cap_i.transpose(1, 2), query=glb_img)
            #     cap_i = cap_t.transpose(1, 2)
            #
            # if glb_txt is not None:
            #     glb_txt_i = glb_txt[i].unsqueeze(0)
            #     img_t = self.adapt_img(value=images.transpose(1, 2), query=glb_txt_i)
            #     img = img_t.transpose(1, 2)
            # else:
            img = images

            if 't2i' in self.sim_name:
                emb_a = cap_i           # (B, T, D)
                emb_b = img             # (B, L, D)
            elif 'i2t' in self.sim_name:
                emb_a = img             # (B, L, D)
                emb_b = cap_i           # (B, T, D)

            if self.max_attn:
                attn = torch.bmm(emb_a, emb_b.transpose(1, 2))
                # (B, T, L) for t2i
                # (B, L, T) for i2t

                dot = torch.max(attn, -1)[0]
                # (B, T) for t2i
                # (B, L) for i2t

            else:
                attn = torch.bmm(emb_a, emb_b.transpose(1, 2))
                # attn /= math.sqrt(emb_a.shape[-1])
                attn = torch.softmax(attn, -1)
                # (B, T, L) for t2i
                # (B, L, T) for i2t

                cxt = torch.bmm(attn, emb_b)
                # (B, T, D) for t2i
                # (B, L, D) for i2t

                cos = torch.nn.CosineSimilarity(dim=-1)
                dot = cos(emb_a, cxt)
                # (B, T) for t2i
                # (B, L) for i2t

            if self.opt.pooling == 'sum':
                dot = dot.sum(dim=-1, keepdim=True)  # (B, 1)
            elif self.opt.pooling == 'mean':
                dot = dot.mean(dim=-1, keepdim=True) # (B, 1)
            similarities.append(dot)

        similarities = torch.cat(similarities, 1)
        return similarities


class MILoss(nn.Module):
    """
    Compute MI loss (approx with mini-batch)
    Working on..
    """

    def __init__(self, opt):
        super(MILoss, self).__init__()
        self.opt = opt

    def forward(self, scores):
        # scores : (B, C) In practice, B=C,
        # but written differently to recognize the modality.

        d_img = d_txt = scores.diag()  # (B,)
        partition_s = scores.logsumexp(0)           # (C) log(sum_(i)^(B)(exp(S(image_i, caption_j))))
        partition_im = scores.logsumexp(1)          # (B) log(sum_(j)^(C)(exp(S(image_i, caption_j))))
        cost_s = - d_txt + partition_s              # (C)
        cost_im = - d_img + partition_im            # (B)

        if self.opt.cost_aggr == 'mean':
            cost_s_t = cost_s.mean()
            cost_im_t = cost_im.mean()
        elif self.opt.cost_aggr == 'sum':
            cost_s_t = cost_s.sum()
            cost_im_t = cost_im.sum()
        else:
            raise ValueError

        return cost_s_t + cost_im_t


class ContrastiveLoss(nn.Module):
    """
    Compute contrastive loss
    """

    def __init__(self, opt):
        super(ContrastiveLoss, self).__init__()

        self.opt = opt
        self.margin = opt.margin
        self.max_violation = opt.max_violation
        self.beta = 0.999
        self.weight = 1.0
        try:
            self.iteration = opt.Eiters
        except:
            self.iteration = 0
        print('Iteration of Loss is set to {}'.format(self.iteration))
        self.k = 0

    def adjust_k(self,):
        self.iteration += 1
        if self.max_violation:
            self.k = 1
        else:
            self.k = (1. - self.beta**np.float(self.iteration))

    def forward(self, scores):
        # scores : (B, C)
        """
        C = B
        (B, 1) -> project to the captions from an image (B, C)
        (1, C) -> project to the images from a caption (B, C)
        """
        diagonal = scores.diag().view(scores.size(0), 1)  # (B, 1)
        d_img = diagonal        # (B, 1)
        d_txt = diagonal.t()    # (1, B)

        d_img = d_img.expand_as(scores)
        d_txt = d_txt.expand_as(scores)

        # compare every diagonal score to scores in its row
        # image retrieval
        cost_im = (self.margin + scores - d_img).clamp(min=0)

        # compare every diagonal score to scores in its column
        # caption retrieval
        cost_s = (self.margin + scores - d_txt).clamp(min=0)
        # clear diagonals
        cost_s.fill_diagonal_(0)
        cost_im.fill_diagonal_(0)

        cost_s_t = cost_s.sum()
        cost_im_t = cost_im.sum()

        if self.training:
            self.adjust_k()

        cost_all_k = (cost_s_t + cost_im_t) * (1. - self.k)
        cost_s_max = cost_s.max(1)[0]
        cost_im_max = cost_im.max(0)[0]

        cost_hard_k = (cost_s_max.sum() + cost_im_max.sum()) * self.k
        total_loss = cost_all_k + cost_hard_k

        return total_loss * self.weight


""" The Architecture of the Visual Graph Encoder """


class EncoderImageSg(nn.Module):
    def __init__(self, tokenizer, embed, fc_word,
                 rel_tokenizer, rel_embed, fc_rel, opt,
                 img_dim, gconv_dim, word_dim,
                 fusion_activation='relu', fusion_method='concatenate',
                 gconv_hidden_dim=1024, gconv_num_layers=5, alpha=1.0,
                 mlp_normalization='none', activation=None,
                 **kwargs):
        super(EncoderImageSg, self).__init__()

        if len(kwargs) > 0:
            print('WARNING: Model got unexpected kwargs ', kwargs)

        self.no_imgnorm = opt.no_imgnorm
        self.no_prednorm = opt.no_prednorm
        self.opt = opt

        # Embedding layers -- BERT
        if opt.shared_model:
            self.embed_dim = opt.embed_size
            self.obj_tokenizer = tokenizer
            self.obj_embed = embed
            self.fc_word = fc_word
            if not opt.no_rel:
                self.rel_tokenizer = rel_tokenizer
                self.rel_embed = rel_embed
                self.fc_rel = fc_rel
        else:
            self.embed_dim = word_dim
            self.obj_tokenizer = get_enc_tokenizer(opt.enc_type)
            self.obj_embed = get_enc_model(opt.enc_type)
            if not opt.no_rel:
                self.rel_tokenizer = get_enc_tokenizer(opt.enc_type)
                self.rel_embed = get_enc_model(opt.enc_type)

        # Multi-modal fusion Layer
        if opt.is_fusion:
            self.obj_fusion = FusionLayer(img_dim, self.embed_dim, fusion_activation, fusion_method)
            if not self.opt.no_rel:
                self.rel_fusion = FusionLayer(img_dim, self.embed_dim, fusion_activation, fusion_method)
            else:
                self.rel_fusion = None
        else:
            self.obj_fusion = None
            self.rel_fusion = None
        # GCN Net
        if self.opt.no_rel:
            self.gconv = nn.Linear(img_dim, gconv_dim)
        else:
            if gconv_num_layers == 0:
                self.gconv = nn.Linear(img_dim, gconv_dim)
                self.gconv_rel = nn.Linear(img_dim, gconv_dim)
            elif gconv_num_layers > 0:
                gconv_kwargs = {
                    'input_dim': img_dim,
                    'output_dim': gconv_dim,
                    'hidden_dim': gconv_hidden_dim,
                    'alpha': alpha,
                    'mlp_normalization': mlp_normalization,
                    'activation': activation,
                }
                self.gconv = GraphTripleConv(**gconv_kwargs)
    
            self.gconv_net = None
            if gconv_num_layers > 1:
                gconv_kwargs = {
                    'input_dim': gconv_dim,
                    'hidden_dim': gconv_hidden_dim,
                    'alpha': alpha,
                    'num_layers': gconv_num_layers - 1,
                    'mlp_normalization': mlp_normalization,
                    'activation': activation
                }
                self.gconv_net = GraphTripleConvNet(**gconv_kwargs)

    def forward(self, obj_embs, pred_embs, rels, rel_labels, obj_labels):
        if self.opt.shared_model:
            with torch.no_grad():
                obj_tokens = self.obj_tokenizer(obj_labels, return_tensors="pt", padding=True)
                obj_tokens = obj_tokens.to('cuda')
                obj_label_embs = self.obj_embed(**obj_tokens).last_hidden_state[:, 1, :]
                obj_label_embs = torch.tanh(self.fc_word(obj_label_embs))
        else:
            obj_tokens = self.obj_tokenizer(obj_labels, return_tensors="pt", padding=True)
            obj_tokens = obj_tokens.to('cuda')
            obj_label_embs = self.obj_embed(**obj_tokens).last_hidden_state[:, 1, :]
        # (batch_size*obj_nums, length, dims) -- select cls embedding --> (batch_size*obj_nums, dims)
        if self.obj_fusion is not None:
            obj_vecs = self.obj_fusion(obj_embs, obj_label_embs)
        else:
            obj_vecs = obj_embs

        if self.opt.no_rel:
            obj_vecs = self.gconv(obj_vecs)
            if not self.no_imgnorm:
                obj_vecs = F.normalize(obj_vecs, p=2, dim=-1, eps=1e-8)
            pred_vecs = None

        else:
            O, T = obj_embs.size(0), rels.size(0)   # (batch_size*obj_nums, dims), (batch_size*pred_nums, dims)
            s, p, o = rels.chunk(3, dim=1)  # All have shape (T, 1)
            s, p, o = [x.squeeze(1) for x in [s, p, o]]  # Now have shape (T,)
            edges = torch.stack([s, o], dim=1)  # Shape is (T, 2)

            if self.opt.shared_model:
                with torch.no_grad():
                    rel_tokens = self.rel_tokenizer(rel_labels, return_tensors="pt", padding=True)
                    rel_tokens = rel_tokens.to('cuda')
                    if self.opt.pred_enc_type == 'SEP':
                        pred_label_embs = self.rel_embed(**rel_tokens).last_hidden_state[:, -1, :]
                    elif self.opt.pred_enc_type == 'last':
                        pred_label_embs = self.rel_embed(**rel_tokens).last_hidden_state[:, -2, :]
                    elif self.opt.pred_enc_type == 'sum':
                        pred_label_embs = self.rel_embed(**rel_tokens).last_hidden_state[:, 1:-1, :].sum(1) # mean(1)
                    else:
                        print("{} is not implemented.".format(self.opt.pred_enc_type))
                    pred_label_embs = torch.tanh(self.fc_rel(pred_label_embs))
            else:
                rel_tokens = self.rel_tokenizer(rel_labels, return_tensors="pt", padding=True)
                rel_tokens = rel_tokens.to('cuda')
                pred_label_embs = self.rel_embed(**rel_tokens).last_hidden_state[:, -1, :]
    
            # pred_label_embs = self.rel_embed(p)     # (T,)1
            if self.rel_fusion is not None:
                pred_vecs = self.rel_fusion(pred_embs, pred_label_embs)
            else:
                pred_vecs = pred_embs
    
            if isinstance(self.gconv, nn.Linear):
                obj_vecs = self.gconv(obj_vecs)
                pred_vecs = self.gconv_rel(pred_vecs)
            else:
                obj_vecs, pred_vecs = self.gconv(obj_vecs, pred_vecs, edges)
            if self.gconv_net is not None:
                obj_vecs, pred_vecs = self.gconv_net(obj_vecs, pred_vecs, edges)
    
            if not self.no_imgnorm:
                obj_vecs = F.normalize(obj_vecs, p=2, dim=-1, eps=1e-8)
            if not self.no_prednorm:
                pred_vecs = F.normalize(pred_vecs, p=2, dim=-1, eps=1e-8)

        return obj_vecs, pred_vecs


""" The Architecture of the Text Graph Encoder """


class EncoderText(nn.Module):

    def __init__(self, tokenizer, embed, fc_word, rel_tokenizer, rel_embed, fc_rel, opt):
        super(EncoderText, self).__init__()
        # word embedding
        self.no_txtnorm = opt.no_txtnorm
        self.no_relnorm = opt.no_relnorm
        self.opt = opt
        if opt.shared_model:
                self.tokenizer = tokenizer
                self.embed = embed
                self.fc_cap = fc_word

                self.cap_rel_tokenizer = rel_tokenizer
                self.cap_rel_embed = rel_embed
                self.fc_cap_rel = fc_rel
        else:
            self.tokenizer = get_enc_tokenizer(opt.enc_type)
            self.embed = get_enc_model(opt.enc_type)
            self.fc_cap = nn.Linear(opt.word_dim, opt.embed_size)

            self.cap_rel_tokenizer = get_enc_tokenizer(opt.enc_type)
            self.cap_rel_embed = get_enc_model(opt.enc_type)
            self.fc_cap_rel = nn.Linear(opt.word_dim, opt.embed_size)

        self.use_cxt = opt.use_cxt
        self.use_pooler_cxt = not opt.no_use_pooler
        self.only_cap = opt.only_cap
        self.enc_type = opt.enc_type

    def forward(self, captions, cap_len, cap_rels, cap_rel_nums):
        tokens = self.tokenizer(captions, return_tensors="pt", padding=True)
        tokens = tokens.to('cuda')
        cap_out = self.embed(**tokens)
        cap_emb = cap_out.last_hidden_state
        cap_emb = torch.tanh(self.fc_cap(cap_emb))
        if not self.no_txtnorm:
            cap_emb = F.normalize(cap_emb, p=2, dim=-1, eps=1e-8)

        if self.opt.no_rel:
            cap_rel_emb = None
        else:
            if self.only_cap:
                tokens = self.cap_rel_tokenizer(captions, return_tensors="pt", padding=True)
                tokens = tokens.to('cuda')
                cap_rel_out = self.cap_rel_embed(**tokens)
                cap_rel_emb = cap_rel_out.last_hidden_state
    
            else:
                tokens = self.cap_rel_tokenizer(cap_rels, return_tensors="pt", padding=True)
                where_sep_tokens = (tokens.input_ids == get_sep_token_id(self.enc_type)).nonzero(as_tuple=False).long()
                tokens = tokens.to('cuda')
                cap_rel_out = self.cap_rel_embed(**tokens)
                cap_rel_emb = cap_rel_out.last_hidden_state
        
                len_sep_tokens = [sum(where_sep_tokens[:, 0] == i).long() for i in range(cap_rel_emb.size(0))]
                new_vectors = torch.zeros(cap_rel_emb.size(0), max(len_sep_tokens).item(), cap_rel_emb.size(2))
                for i in range(cap_rel_emb.size(0)):
                    if self.opt.pred_enc_type == 'SEP':
                        new_vectors[i, :len_sep_tokens[i]] = cap_rel_emb[i][where_sep_tokens[where_sep_tokens[:, 0] == i][:, 1]]
                    elif self.opt.pred_enc_type == 'last':
                        new_vectors[i, :len_sep_tokens[i]] = cap_rel_emb[i][where_sep_tokens[where_sep_tokens[:, 0] == i][:, 1]-1]
                    elif self.opt.pred_enc_type == 'sum':
                        where_sep_end = where_sep_tokens[where_sep_tokens[:, 0] == i][:, 1]
                        where_sep_start = torch.cat((torch.zeros(1, dtype=int), where_sep_end[:-1]))+1
                        for j in range(len_sep_tokens[i]):
                            new_vector = cap_rel_emb[i, where_sep_start[j]:where_sep_end[j], :]
                            new_vectors[i, j] = new_vector.sum(0)
                    else:
                        print("Not implemented {}".format(self.opt.pred_enc_type))
        
                cap_rel_emb = new_vectors.cuda()
                # (batch, cap_rel_nums, dims)
                if self.use_cxt:
                    if self.use_pooler_cxt:
                        cxt = cap_out.pooler_output.unsqueeze(1)
                    else:
                        cxt = cap_out.last_hidden_state.mean(1, keepdim=True)
                    cap_rel_emb = torch.cat((cxt, cap_rel_emb), 1)
                    cap_rel_nums = [(n+1) for n in cap_rel_nums]
    
            cap_rel_emb = torch.tanh(self.fc_cap_rel(cap_rel_emb))
            if not self.no_relnorm:
                cap_rel_emb = F.normalize(cap_rel_emb, p=2, dim=-1, eps=1e-8)

        return cap_emb, cap_len, cap_rel_emb, cap_rel_nums


class RELAX(nn.Module):
    def __init__(self, opt):
        super(RELAX, self).__init__()
        # Build Models
        self.opt = opt
        self.grad_clip = opt.grad_clip

        self.tokenizer = None
        self.embed = None
        self.fc_word = None

        self.rel_tokenizer = None
        self.rel_embed = None
        self.fc_rel = None
        if opt.shared_model:
            self.tokenizer = get_enc_tokenizer(opt.enc_type)
            self.embed = get_enc_model(opt.enc_type)
            self.fc_word = nn.Linear(opt.word_dim, opt.embed_size)
            if not opt.no_rel:
                self.rel_tokenizer = get_enc_tokenizer(opt.enc_type)
                self.rel_embed = get_enc_model(opt.enc_type)
                self.fc_rel = nn.Linear(opt.word_dim, opt.embed_size)

        self.img_enc = EncoderImageSg(self.tokenizer, self.embed, self.fc_word,
                                      self.rel_tokenizer, self.rel_embed, self.fc_rel, opt,
                                      opt.img_dim, opt.embed_size, opt.word_dim,
                                      fusion_activation=opt.fusion_activation, fusion_method=opt.fusion_method,
                                      gconv_hidden_dim=1024, gconv_num_layers=opt.gcn_num_layers,
                                      alpha=opt.alpha,
                                      activation=opt.activation)

        self.txt_enc = EncoderText(self.tokenizer, self.embed, self.fc_word,
                                   self.rel_tokenizer, self.rel_embed, self.fc_rel,
                                   opt=opt)
        self.adapt_img = self.adapt_txt = self.rel_adapt_img = self.rel_adapt_txt = False
        if opt.adapt:
            self.rel_adapt_img = True
            if opt.cross_attn == 't2i':
                self.adapt_img = True
            if opt.cross_attn == 'i2t':
                self.adapt_txt = True
        self.similarity = Similarity(AdaptiveEmbedding, opt, opt.cross_attn, opt.max_attn)
        if not opt.no_rel:
            self.rel_sim = Similarity(AdaptiveEmbedding, opt, opt.rel_cross_attn, opt.max_attn)
            
        self.init_inv_t = torch.zeros(1).fill_(opt.inv_t).cuda()  # .cuda().requires_grad_(True)
        self.inv_t = self.init_inv_t
        if torch.cuda.is_available():
            self.img_enc.cuda()
            self.txt_enc.cuda()
            self.similarity.cuda()
            if not opt.no_rel:
                self.rel_sim.cuda()
            cudnn.benchmark = True

        # Loss and Optimizer
        if 'margin' in opt.loss_type:
            self.criterion = ContrastiveLoss(opt=opt)
        else:  # MI
            self.criterion = MILoss(opt=opt)

        params = list(self.txt_enc.parameters())
        params += list(self.img_enc.parameters())
        params += list(self.similarity.parameters())
        if not self.opt.no_rel:
            params += list(self.rel_sim.parameters())
        self.params = params

        self.optimizer = torch.optim.Adam(params, lr=opt.learning_rate)

        self.Eiters = 0
        self.moving_L = []
        self.moving_avg_L = None
        self.gamma = 0.1
 
    def forward(self, images, obj_nums, captions, lengths,
                image_predicates, pred_nums, rels, objs, cap_rels, cap_rel_nums,
                obj_labels, rel_labels,
                max_obj_n=36, max_pred_n=25):

        if torch.cuda.is_available():
            images = images.cuda()
            objs = objs.cuda()
            if not self.opt.no_rel:
                image_predicates = image_predicates.cuda()
                rels = rels.cuda()

        img_vecs, pred_vecs = self.img_enc(images, image_predicates, rels, rel_labels, obj_labels)
        cap_emb, cap_lens, cap_rel_emb, cap_rel_nums = self.txt_enc(captions, lengths, cap_rels, cap_rel_nums)

        img_emb = torch.zeros(len(obj_nums), max_obj_n, img_vecs.shape[1]).cuda()

        obj_offset = 0
        for i, obj_num in enumerate(obj_nums):
            img_emb[i][:obj_num] = img_vecs[obj_offset:obj_offset + obj_num]
            obj_offset += obj_num

        if not self.opt.no_rel:
            pred_emb = torch.zeros(len(pred_nums), max_pred_n, pred_vecs.shape[1]).cuda()
        
            pred_offset = 0
            for i, pred_num in enumerate(pred_nums):
                pred_emb[i][:pred_num] = pred_vecs[pred_offset: pred_offset + pred_num]
                pred_offset += pred_num
        else:
            pred_emb = None

        return img_emb, cap_emb, cap_lens, pred_emb, cap_rel_emb, cap_rel_nums

    def forward_loss(self, scores):

        """Compute the loss given pairs of image and caption embeddings
        """
        loss = self.criterion(scores)
        self.logger.update('Le', loss.item(), scores.size(0))
        if 'margin' in self.opt.loss_type:
            self.logger.update('K', self.criterion.k)
        else:
            self.logger.update('S', scores.sum().item())
            self.logger.update('1/T', self.inv_t[0].item())
        return loss

    def get_global(self, img_emb, cap_emb, adapt_img, adapt_txt, use_cxt=False):
        if adapt_txt:      # for i2t in ADAPT
            glb_img = img_emb.mean(1)
        else:
            glb_img = None

        if adapt_img:      # for t2i in ADAPT
            if use_cxt:
                glb_txt = cap_emb[:, 0]
            else:
                glb_txt = cap_emb.mean(1)
        else:
            glb_txt = None

        return glb_img, glb_txt

    def train_emb(self, images, obj_nums, captions, lengths, ids, image_predicates, pred_nums, rels, objs, cap_rels,
                  cap_rel_nums, obj_labels, rel_labels):

        self.Eiters += 1
        self.logger.update('Eit', self.Eiters)
        self.logger.update('lr', self.optimizer.param_groups[0]['lr'])

        # compute the embeddings
        img_emb, cap_emb, cap_lens, pred_emb, cap_rel_emb, cap_rel_lens = \
            self.forward(images, obj_nums, captions, lengths, image_predicates, pred_nums, rels, objs, cap_rels,
                         cap_rel_nums, obj_labels, rel_labels)

        glb_img, glb_txt = self.get_global(img_emb, cap_emb, self.adapt_img, self.adapt_txt, use_cxt=False)
        scores1 = self.similarity(img_emb, cap_emb, cap_lens,
                                  glb_img=glb_img, glb_txt=glb_txt,
                                  exclude_last=True)
        if self.opt.no_rel:
            scores2 = torch.zeros_like(scores1)
        else:
            glb_rel_img, glb_rel_txt = self.get_global(pred_emb, cap_rel_emb, self.rel_adapt_img, self.rel_adapt_txt, use_cxt=True)
            scores2 = self.rel_sim(pred_emb, cap_rel_emb, cap_rel_lens,
                               glb_img=glb_rel_img, glb_txt=glb_rel_txt)
        scores = scores1 + self.opt.predicate_score_rate * scores2
        
        if 'mi' in self.opt.loss_type:
            scores *= self.inv_t

        self.optimizer.zero_grad()
        loss = self.forward_loss(scores)

        # compute gradient and do SGD step
        loss.backward()
        if self.grad_clip > 0:
            clip_grad_norm(self.params, self.grad_clip)
        self.optimizer.step()
