# -----------------------------------------------------------
# "Image-Text Alignment using Adaptive Cross-attention with Transformer Encoder for Scene Graphs"
# Juyong Song and Sunghyun Choi
#
# Modified by Juyong Song, 2021
# ---------------------------------------------------------------

# -----------------------------------------------------------
# "Cross-modal Scene Graph Matching for Relationship-aware Image-Text Retrieval"
# Sijin Wang, Ruiping Wang, Ziwei yao, Shiguang Shan and Xilin Chen
# https://arxiv.org/abs/1910.05134
#
# Written by Sijin Wang, 2019
# ---------------------------------------------------------------

# -----------------------------------------------------------
# Stacked Cross Attention Network implementation based on
# https://arxiv.org/abs/1803.08024.
# "Stacked Cross Attention for Image-Text Matching"
# Kuang-Huei Lee, Xi Chen, Gang Hua, Houdong Hu, Xiaodong He
#
# Written by Kuang-Huei Lee, 2018
# ---------------------------------------------------------------

"""Training script"""

import sys
import argparse
import logging
import os
import random
import shutil
import time
import math
import copy

import numpy
import tensorboard_logger as tb_logger
import torch

import data
from evaluation import i2t, t2i, AverageMeter, LogCollector, \
    encode_data_with_predicate
from model import RELAX

train_logger = LogCollector()
logger = logging.getLogger(__name__)
logger.setLevel(level=logging.DEBUG)


def main():
    # Hyper Parameters
    parser = argparse.ArgumentParser()
    parser.add_argument('--data_path', default='./data/',
                        help='path to datasets')
    parser.add_argument('--cache_path', default='./data_cache/',
                        help='path to cache of scene graphs data')
    parser.add_argument('--data_name', default='flickr30k_split',
                        help='flickr30k_split')
    parser.add_argument('--margin', default=0.2, type=float,
                        help='Rank loss margin.')
    parser.add_argument('--num_epochs', default=30, type=int,
                        help='Number of training epochs.')
    parser.add_argument('--batch_size', default=128, type=int,
                        help='Size of a training mini-batch.')
    parser.add_argument('--word_dim', default=768, type=int,
                        help='Dimensionality of the word embedding.')
    parser.add_argument('--embed_size', default=1024, type=int,
                        help='Dimensionality of the joint embedding.')
    parser.add_argument('--grad_clip', default=2., type=float,
                        help='Gradient clipping threshold.')
    parser.add_argument('--num_layers', default=1, type=int,
                        help='Number of GRU layers.')
    parser.add_argument('--gcn_num_layers', default=1, type=int,
                        help='Number of image sg gcn layers')
    parser.add_argument('--activation', default='tanh', type=str,
                        help='activate function use in gcn')
    parser.add_argument('--learning_rate', default=2.5e-5, type=float,
                        help='Initial learning rate.')
    parser.add_argument('--lr_update', default=15, type=int,
                        help='Number of epochs to update the learning rate.')
    parser.add_argument('--workers', default=0, type=int,
                        help='Number of data loader workers.')
    parser.add_argument('--log_step', default=10, type=int,
                        help='Number of steps to print and record the log.')
    parser.add_argument('--val_step', default=500, type=int,
                        help='Number of steps to run validation.')
    parser.add_argument('--logger_name', default='./log',
                        help='Path to save Tensorboard log.')
    parser.add_argument('--model_name', default='./checkpoint',
                        help='Path to save the model.')
    parser.add_argument('--resume', default=None, type=str, metavar='PATH',
                        help='path to latest checkpoint (default: none)')
    parser.add_argument('--max_violation', action='store_true',
                        help='Use max instead of sum in the rank loss.')
    parser.add_argument('--img_dim', default=2048, type=int,
                        help='Dimensionality of the image embedding.')
    parser.add_argument('--no_imgnorm', action='store_true',
                        help='Do not normalize the image embeddings.')
    parser.add_argument('--no_txtnorm', action='store_true',
                        help='Do not normalize the text embeddings.')
    parser.add_argument('--no_prednorm', action='store_true',
                        help='Do not normalize the image predicate embeddings.')
    parser.add_argument('--no_relnorm', action='store_true',
                        help='Do not normalize the text relation embeddings.')
    parser.add_argument('--predicate_score_rate', type=float, default=1.0,
                        help ='the rate of the relationship scores')
    parser.add_argument('--cross_attn', default='t2i',
                        help='t2i|i2t')
    parser.add_argument('--rel_cross_attn', default='t2i',
                        help='t2i|i2t')
    parser.add_argument('--debug', action='store_true', 
                        help='If in the debug mode, load the validation data')
    parser.add_argument('--sg_file_name', type =str, default='_image_sg_by_NeuralMotifs_36_25.json',
                        help='The file that contains the visual scene graph information')
    parser.add_argument('--feature_name', type=str, default='BUA',
                        help='The features come from which detector')
    parser.add_argument('--caption_data_file_name', type=str, default='_caps_with_rel.json',
                        help='The file that contains the textual scene graph information')
    parser.add_argument('--is_fusion', action='store_true',
                        help='If fuse the predicted label with the visual features or not. Default : False')
    parser.add_argument('--alpha', type=float, default=1.0)
    parser.add_argument('--fusion_activation', type=str, default='tanh',
                        help='The activation function of the fusion layer.')
    parser.add_argument('--fusion_method', type=str, default='concatenate',
                        help='The fusion method of the fusion layer.')
    parser.add_argument('--save_all', action='store_true', help='Full save at each validation Default : False')
    parser.add_argument('--seed', default=42, type=int,
                        help='Seed of randomness.')
    parser.add_argument('--part_of_data', default=None, type=float,
                        help='Use the part of data. The value is between 0 and 1.')
    parser.add_argument('--max_attn', action='store_true', help='MAX attention used in SGM.')

    parser.add_argument('--adapt', action='store_true', help='Adaptive embedding for key and value')
    parser.add_argument('--inv_t', default=10, type=float, help='Init temperature for MI.')
    parser.add_argument('--enc_type', default='bert_base', choices=['bert_base', 'bert_large', 'albert_base',
                                                                    'albert_large', 'albert_xlarge', 'albert_xxlarge',
                                                                    'roberta_base', 'distilbert_base'])
    parser.add_argument('--loss_type', choices=['margin', 'mi'], default='mi',
                        help='Loss type, we support contrastive loss, mutual information loss')
    parser.add_argument('--only_cap', action='store_true',
                        help='Use cap rel. Default False')
    parser.add_argument('--is_triplet', action='store_true',
                        help='If use the predicate label as the triplets including the object labels. Default : False')
    parser.add_argument('--use_cxt', action='store_true',
                        help='If add the caption pooler embedding to cap_rel_emb. Default : False')
    parser.add_argument('--no_use_pooler', action='store_true',
                        help='If add the caption mean to cap_rel_emb. Default : False (pooler)')
    parser.add_argument('--shared_model', action='store_true',
                        help='Model sharing for the labeling of the image scene graphs and caption')
    parser.add_argument('--pred_enc_type', choices=['SEP', 'last', 'sum'], default='SEP',
                        help='Predicate or relation BERT encoding type.')
    parser.add_argument('--cost_aggr', choices=['mean', 'sum'], default='sum')
    parser.add_argument('--pooling', choices=['mean', 'sum'], default='sum')
    parser.add_argument('--no_rel', action='store_true')

    opt = parser.parse_args()
    if opt.part_of_data is not None:
        opt.num_epochs = int(math.ceil(opt.num_epochs / opt.part_of_data))
        opt.lr_update = int(math.ceil(opt.lr_update / opt.part_of_data))

    os.makedirs(opt.model_name, exist_ok=True)
    os.makedirs(opt.logger_name, exist_ok=True)
    torch.manual_seed(opt.seed)
    random.seed(opt.seed, version=2)
    numpy.random.seed(seed=opt.seed)

    log_path = os.path.join(opt.logger_name,'out.log')

    formatter = logging.Formatter('%(asctime)s %(message)s')
    streamHandler = logging.StreamHandler(sys.stdout)
    fileHandler = logging.FileHandler(log_path)
    streamHandler.setFormatter(formatter)
    fileHandler.setFormatter(formatter)

    logger.addHandler(streamHandler)
    logger.addHandler(fileHandler)
    logger.info('{}'.format(opt))
    
    tb_logger.configure(opt.logger_name, flush_secs=5)

    """ Load data """
    print("loading data")

    # Get the train and val config filenames
    train_file_list = [opt.data_name.replace('split', 'train') + opt.sg_file_name, 
                           opt.data_name.replace('split', 'train') + opt.caption_data_file_name]

    val_file_list = [opt.data_name.replace('split', 'val') + opt.sg_file_name, 
                         opt.data_name.replace('split', 'val') + opt.caption_data_file_name]

    # file_list[0] : SG json (dict)
    # file_list[1] : caption json (list)

    # Load the training data and validation data and the image object/relation vocab
    train_loader, val_loader = data.get_loaders(opt.data_name, train_file_list, val_file_list, opt.feature_name, opt.batch_size, opt.workers, opt)

    print("Data loaded")

    """ Construct the model """
    best_rsum = 0
    best_TRsum = 0
    best_IRsum = 0
    # optionally resume from a checkpoint
    if opt.resume is not None:
        if os.path.isfile(opt.resume):
            print("=> loading checkpoint '{}'".format(opt.resume))
            checkpoint = torch.load(opt.resume)
            for k, v in checkpoint.items():
                if k == 'model':
                    continue
                print(k, v)
            opt_resume = checkpoint['opt']
            print(opt_resume)
            opt_resume.logger_name = opt.logger_name
            opt_resume.model_name = opt.model_name
            opt_resume.resume = opt.resume
            opt_resume.val_step = opt.val_step
            opt_resume.lr_update = opt.lr_update
            opt_resume.num_epochs = opt.num_epochs
            opt_resume.max_violation = opt.max_violation
            opt = copy.deepcopy(opt_resume)

            opt.Eiters = checkpoint['Eiters']
            logger.info('{}'.format(opt))
            # opt.batch_size=10 # for a local run
            model = RELAX(opt)
            start_epoch = checkpoint['epoch']-1
            best_rsum = checkpoint['best_rsum']
            model.load_state_dict(checkpoint['model'])
            # Eiters is used to show logs as the continuation of another
            # training
            model.Eiters = checkpoint['Eiters']
            try:
                best_TRsum = checkpoint['best_TRsum']
                best_IRsum = checkpoint['best_IRsum']
                print("best_TRsum:{} best_IRsum:{}".format(best_TRsum, best_IRsum))
            except:
                print("No TR, IR best sum")
            print("=> loaded checkpoint '{}' (epoch {}, best_rsum {})"
                  .format(opt.resume, start_epoch, best_rsum))
            rsum, tr_sum, ir_sum = validate(opt, val_loader, model)
            best_rsum = max(rsum, best_rsum)
            best_TRsum = max(tr_sum, best_TRsum)
            best_IRsum = max(ir_sum, best_IRsum)
        else:
            print("=> no checkpoint found at '{}'".format(opt.resume))
            model = RELAX(opt)
            start_epoch = 0
    else:
       model = RELAX(opt)
       start_epoch = 0

    """ Start training """
    for epoch in range(start_epoch, opt.num_epochs):
        adjust_learning_rate(opt, model.optimizer, epoch)
        # train for one epoch
        best_rsum, best_TRsum, best_IRsum = train(
            opt, train_loader, model, epoch, val_loader, best_rsum, best_TRsum, best_IRsum)

        # evaluate on validation set
        rsum, tr_sum, ir_sum = validate(opt, val_loader, model)

        # remember best R@ sum and save checkpoint
        best_rsum, best_TRsum, best_IRsum = call_save_checkpoint(
            opt, model, epoch, rsum, tr_sum, ir_sum, best_rsum, best_TRsum, best_IRsum)
 

def call_save_checkpoint(opt, model, epoch, rsum, tr_sum, ir_sum, best_rsum, best_TRsum, best_IRsum):
    print(f"rsum = {rsum}, best_rsum = {best_rsum}")
    print(f"tr_sum = {tr_sum}, best_TRsum = {best_TRsum}")
    print(f"ir_sum = {ir_sum}, best_IRsum = {best_IRsum}")

    is_best = rsum > best_rsum
    is_TRbest = tr_sum > best_TRsum
    is_IRbest = ir_sum > best_IRsum
    print(f"is_best : {is_best}, is_TRbest : {is_TRbest}, is_IRbest : {is_IRbest}")

    best_rsum = max(rsum, best_rsum)
    best_TRsum = max(tr_sum, best_TRsum)
    best_IRsum = max(ir_sum, best_IRsum)

    if opt.save_all:
        filename = 'checkpoint_{}_{}.pth.tar'.format(epoch, model.Eiters)
    else:
        filename = 'checkpoint_last.pth.tar'
    save_checkpoint({
        'epoch': epoch + 1,
        'model': model.state_dict(),
        'best_rsum': best_rsum,
        'best_IR': best_IRsum,
        'best_TR': best_TRsum,
        'opt': opt,
        'Eiters': model.Eiters,
    }, is_best, is_TRbest, is_IRbest, filename=filename, prefix=opt.model_name + '/')

    return best_rsum, best_TRsum, best_IRsum


def train(opt, train_loader, model, epoch, val_loader, best_rsum, best_TRsum, best_IRsum):
    # average meters to record the training statistics
    batch_time = AverageMeter()
    data_time = AverageMeter()

    end = time.time()
    # switch to train mode
    model.train()
    for i, train_data in enumerate(train_loader):

        # measure data loading time
        data_time.update(time.time() - end)

        # make sure train logger is used
        model.logger = train_logger

        # Update the model_name
        model.train_emb(*train_data)

        # measure elapsed time
        batch_time.update(time.time() - end)
        end = time.time()

        # Print log info
        if model.Eiters % opt.log_step == 0:
            logger.info(
                'Epoch: [{0}][{1}/{2}]\t'
                '{e_log}\t'
                'Time {batch_time.val:.3f} ({batch_time.avg:.3f})\t'
                'Data {data_time.val:.3f} ({data_time.avg:.3f})\t'
                .format(
                    epoch, i, len(train_loader), batch_time=batch_time,
                    data_time=data_time, e_log=str(model.logger)))

        # Record logs in tensorboard
        tb_logger.log_value('epoch', epoch, step=model.Eiters)
        tb_logger.log_value('step', i, step=model.Eiters)
        tb_logger.log_value('batch_time', batch_time.val, step=model.Eiters)
        tb_logger.log_value('data_time', data_time.val, step=model.Eiters)
        model.logger.tb_log(tb_logger, step=model.Eiters)

        # validate at every val_step
        if model.Eiters % opt.val_step == 0:
            rsum, tr_sum, ir_sum = validate(opt, val_loader, model)
            # remember best R@ sum and save checkpoint
            best_rsum, best_TRsum, best_IRsum = call_save_checkpoint(
                    opt, model, epoch, rsum, tr_sum, ir_sum, best_rsum, best_TRsum, best_IRsum)
            # switch to train mode again
            model.train()

    return best_rsum, best_TRsum, best_IRsum


def validate(opt, val_loader, model):
    # compute the encoding for all the validation images and captions
    model.eval() # val_start()
    with torch.no_grad():
        img_embs, obj_nums, cap_embs, cap_lens, pred_embs, pred_nums, cap_rel_embs, cap_rel_nums = encode_data_with_predicate(
            model, val_loader, opt, opt.log_step, logger.info)

        img_embs = torch.FloatTensor([img_embs[i] for i in range(0, len(img_embs), 5)]).cuda()
        cap_embs = torch.FloatTensor(cap_embs).cuda()
        
        if not opt.no_rel:
            pred_embs = torch.FloatTensor([pred_embs[i] for i in range(0, len(pred_embs), 5)]).cuda()
            cap_rel_embs = torch.FloatTensor(cap_rel_embs).cuda()

        start = time.time()

        glb_img, glb_txt = model.get_global(img_embs, cap_embs, model.adapt_img, model.adapt_txt, use_cxt=False)
        sims1 = model.similarity.forward_shared(img_embs, cap_embs, cap_lens,
                                                glb_img=glb_img, glb_txt=glb_txt,
                                                shared_size=128, exclude_last=True)
        if opt.no_rel:
            sims2 = torch.zeros_like(sims1)
        else:
            glb_rel_img, glb_rel_txt = model.get_global(pred_embs, cap_rel_embs, model.rel_adapt_img, model.rel_adapt_txt, use_cxt=True)
    
            sims2 = model.rel_sim.forward_shared(pred_embs, cap_rel_embs, cap_rel_nums,
                                                 glb_img=glb_rel_img, glb_txt=glb_rel_txt,
                                                 shared_size=128)

        sims = sims1 + opt.predicate_score_rate * sims2
        
        end = time.time()
        print("calculate similarity time:", end-start)

    sims = sims.data.cpu().numpy()
    # caption retrieval
    (r1, r5, r10, medr, meanr),_ = i2t(img_embs, cap_embs, cap_lens, sims)
    logger.info("Image to text: %.1f, %.1f, %.1f, %.1f, %.1f" %
                 (r1, r5, r10, medr, meanr))
    # image retrieval
    (r1i, r5i, r10i, medri, meanri),_ = t2i(
        img_embs, cap_embs, cap_lens, sims)
    logger.info("Text to image: %.1f, %.1f, %.1f, %.1f, %.1f" %
                 (r1i, r5i, r10i, medri, meanri))
    # sum of recalls to be used for early stopping
    currscore = r1 + r5 + r10 + r1i + r5i + r10i
    tr_sum = r1+r5+r10
    ir_sum = r1i+r5i+r10i
    logger.info("rsum: %.1f (TR) %.1f (IR) %.1f" % (currscore, tr_sum, ir_sum))

    # record metrics in tensorboard
    tb_logger.log_value('r1', r1, step=model.Eiters)
    tb_logger.log_value('r5', r5, step=model.Eiters)
    tb_logger.log_value('r10', r10, step=model.Eiters)
    tb_logger.log_value('medr', medr, step=model.Eiters)
    tb_logger.log_value('meanr', meanr, step=model.Eiters)
    tb_logger.log_value('tr_sum', tr_sum, step=model.Eiters)
    tb_logger.log_value('r1i', r1i, step=model.Eiters)
    tb_logger.log_value('r5i', r5i, step=model.Eiters)
    tb_logger.log_value('r10i', r10i, step=model.Eiters)
    tb_logger.log_value('medri', medri, step=model.Eiters)
    tb_logger.log_value('meanr', meanri, step=model.Eiters)
    tb_logger.log_value('ir_sum', ir_sum, step=model.Eiters)
    tb_logger.log_value('rsum', currscore, step=model.Eiters)

    return currscore, tr_sum, ir_sum


def save_checkpoint(state, is_best, is_tr_best, is_ir_best, filename='checkpoint.pth.tar', prefix=''):
    tries = 15
    error = None

    # deal with unstable I/O. Usually not necessary.
    while tries:
        try:
            torch.save(state, prefix + filename)
            if is_best:
                shutil.copyfile(prefix + filename, prefix + 'model_best.pth.tar')
#             if is_tr_best:
#                 shutil.copyfile(prefix + filename, prefix + 'model_TRbest.pth.tar')
#             if is_ir_best:
#                 shutil.copyfile(prefix + filename, prefix + 'model_IRbest.pth.tar')
        except IOError as e:
            error = e
            tries -= 1
        else:
            break
        print('model save {} failed, remaining {} trials'.format(filename, tries))
        if not tries:
            raise error


def adjust_learning_rate(opt, optimizer, epoch):
    """Sets the learning rate to the initial LR
       decayed by 10 every lr_update epochs"""
    lr = opt.learning_rate * (0.1 ** (epoch // opt.lr_update))
    for param_group in optimizer.param_groups:
        param_group['lr'] = lr


def accuracy(output, target, topk=(1,)):
    """Computes the precision@k for the specified values of k"""
    maxk = max(topk)
    batch_size = target.size(0)

    _, pred = output.topk(maxk, 1, True, True)
    pred = pred.t()
    correct = pred.eq(target.view(1, -1).expand_as(pred))

    res = []
    for k in topk:
        correct_k = correct[:k].view(-1).float().sum(0)
        res.append(correct_k.mul_(100.0 / batch_size))
    return res


if __name__ == '__main__':
    main()
